@@ -15,8 +15,8 @@
  *
  * You should have received a copy of the GNU Library General Public
  * License along with this library; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA 02111-1307, USA.
+ * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
+ * Boston, MA 02110-1301, USA.
  */
 
 /**
@@ -101,11 +101,7 @@ gupnp_device_info_set_property (GObject      *object,
                 info->priv->device_type = g_value_dup_string (value);
                 break;
         case PROP_URL_BASE:
-                info->priv->url_base = g_value_get_pointer (value);
-                if (info->priv->url_base)
-                        info->priv->url_base =
-                                soup_uri_copy (info->priv->url_base);
-
+                info->priv->url_base = g_value_dup_boxed (value);
                 break;
         case PROP_DOCUMENT:
                 info->priv->doc = g_value_dup_object (value);
@@ -151,8 +147,7 @@ gupnp_device_info_get_property (GObject    *object,
                                     gupnp_device_info_get_device_type (info));
                 break;
         case PROP_URL_BASE:
-                g_value_set_pointer (value,
-                                     info->priv->url_base);
+                g_value_set_boxed (value, info->priv->url_base);
                 break;
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
@@ -314,14 +309,15 @@ gupnp_device_info_class_init (GUPnPDeviceInfoClass *klass)
         g_object_class_install_property
                 (object_class,
                  PROP_URL_BASE,
-                 g_param_spec_pointer ("url-base",
-                                       "URL base",
-                                       "The URL base",
-                                       G_PARAM_READWRITE |
-                                       G_PARAM_CONSTRUCT_ONLY |
-                                       G_PARAM_STATIC_NAME |
-                                       G_PARAM_STATIC_NICK |
-                                       G_PARAM_STATIC_BLURB));
+                 g_param_spec_boxed ("url-base",
+                                     "URL base",
+                                     "The URL base",
+                                     SOUP_TYPE_URI,
+                                     G_PARAM_READWRITE |
+                                     G_PARAM_CONSTRUCT_ONLY |
+                                     G_PARAM_STATIC_NAME |
+                                     G_PARAM_STATIC_NICK |
+                                     G_PARAM_STATIC_BLURB));
 
         /**
          * GUPnPDeviceInfo:document:
@@ -1109,7 +1105,7 @@ gupnp_device_info_list_device_types (GUPnPDeviceInfo *info)
  * this function a new object is created. The application must cache any used
  * devices if it wishes to keep them around and re-use them.
  *
- * Returns: (transfer full): A new #GUPnPDeviceInfo.
+ * Returns: (transfer full)(allow-none): A new #GUPnPDeviceInfo.
  **/
 GUPnPDeviceInfo *
 gupnp_device_info_get_device (GUPnPDeviceInfo *info,
